'
'  BaseGA.vb - Base class for simple genetic algorithm implementation.
'
'  Project: Framework
'  Author: S.Zabinskis
'  December, 2006
'
Imports System.Math
Imports System.Collections.Generic


Namespace GeneticAlgorithm

    Public MustInherit Class BaseGeneticAlgorithm(Of T As IChromosome)

#Region "Local Types"
        Public Class ValuedObject(Of X)
            Public _value As Double
            Public _rndval As Double
            Public _object As X
            Public Sub New(ByVal obj As X, ByVal value As Double)
                _object = obj
                _value = value
            End Sub
        End Class

        Private Class Mixer : Implements IComparer(Of ValuedObject(Of T))

            Public Function Compare(ByVal x As ValuedObject(Of T), ByVal y As ValuedObject(Of T)) As Integer Implements System.Collections.Generic.IComparer(Of ValuedObject(Of T)).Compare
                If x._rndval = y._rndval Then
                    Return 0
                End If
                Return IIf(x._rndval > y._rndval, 1, -1)
            End Function
        End Class

#End Region

#Region "Local Variables"
        Protected Shared _random As Random = New Random((DateTime.Now.Hour * 3600 + DateTime.Now.Minute * 60 + DateTime.Now.Second) * 1000 + DateTime.Now.Millisecond)
        Protected _population As New List(Of ValuedObject(Of T))
        Protected _iteration As Integer = 0
#End Region

        '#Region "Constructors"
        '        Public Sub New()
        '        End Sub
        '#End Region

#Region "Abstract methods"
        Public MustOverride Function EvalFitness(ByVal obj As T) As Double
        Protected MustOverride Function CanStop(ByVal iteration As Integer, ByVal population As List(Of T)) As Boolean
        Protected MustOverride Function PerformMutate() As Boolean
        Protected MustOverride Function PerformCrossover() As Boolean
        Public MustOverride Function CreateChromosome() As T
#End Region

#Region "Methods"

        Protected Function GetFitness(ByVal obj As T) As Double
            For Each vo As ValuedObject(Of T) In _population
                If vo._object.Equals(obj) Then
                    Return vo._value
                End If
            Next
            Return EvalFitness(obj)
        End Function

        Private Function NormalizePopulation(ByVal population As IEnumerable(Of T), ByVal reverse_flag As Boolean) As Boolean
            Dim minftn As Double = Double.MaxValue
            Dim maxftn As Double = Double.MinValue
            Dim sumftn As Double = 0
            Dim fitness As Double = Double.NaN

            Dim N As Integer = 0

            'Debug.WriteLine("NormalizePopulation")
            _population.Clear()

            For Each obj As T In population
                If reverse_flag Then
                    fitness = -EvalFitness(obj)
                Else
                    fitness = EvalFitness(obj)
                End If
                If maxftn < fitness Then
                    maxftn = fitness
                End If
                If minftn > fitness Then
                    minftn = fitness
                End If
                sumftn += fitness
                Dim vo As New ValuedObject(Of T)(obj, fitness)
                vo._rndval = BaseGeneticAlgorithm(Of T)._random.NextDouble()
                _population.Add(vo)
                N += 1
                'Debug.WriteLine(String.Format("filtness={0}", fitness))
            Next
            sumftn -= minftn * N

            Dim normf As Double = 1.0 / sumftn
            For Each vo As ValuedObject(Of T) In _population
                vo._value = (vo._value - minftn) * normf
            Next
            'Debug.WriteLine("end of NormalizePopulation")
            Return N > 1
        End Function

        Protected Overridable Sub ShufflePopulation()
            _population.Sort(New Mixer())
        End Sub


        Private Function Pick(ByVal population As List(Of ValuedObject(Of T))) As T
            Dim p As Double = _random.NextDouble
            Dim sum As Double = 0
            Dim o As T = Nothing

            For Each obj As ValuedObject(Of T) In population
                sum += obj._value
                If sum >= p Then
                    Return obj._object
                End If
                o = obj._object
            Next
            Return o
        End Function


        Protected Sub Mutate(ByVal obj As T)
            Dim locus As Integer = 0
            While obj.ValidLocus(locus)
                If PerformMutate() Then
                    obj.MutateAt(locus)
                End If
                locus += 1
            End While
        End Sub

        Private Function PickBest(ByVal population As List(Of ValuedObject(Of T))) As T
            Dim maxf As Double = Double.MinValue
            Dim o As T = Nothing
            For Each obj As ValuedObject(Of T) In population
                If obj._value > maxf Then
                    maxf = obj._value
                    o = obj._object
                End If
            Next
            Return o
        End Function

        Private Function NextPopulation(ByVal reverse_flag As Boolean) As List(Of T)

            Dim newpopulation As New List(Of T)
            Dim N As Integer = _population.Count
            Dim counter As Integer = 0

            ShufflePopulation()

            ' elitism
            Dim bestObj As T = PickBest(_population)
            newpopulation.Add(bestObj)
            Dim offspring1 As T = Nothing
            Dim offspring2 As T = Nothing


            While counter < N

                Dim parent1 As T = Pick(_population)
                Dim parent2 As T = Pick(_population)
                While parent1.Equals(parent2)
                    parent2 = Pick(_population)
                End While

                offspring1 = CreateChromosome()
                offspring2 = CreateChromosome()

                If PerformCrossover() Then
                    parent1.CrossOver(parent2, offspring1, offspring2)
                Else
                    parent1.Copy(offspring1)
                    parent1.Copy(offspring2)
                End If
                Mutate(offspring1)
                Mutate(offspring2)
                newpopulation.Add(offspring1)
                newpopulation.Add(offspring2)
                counter += 2

                While counter > N
                    newpopulation.RemoveAt(counter - 1)
                    counter -= 1
                End While
            End While

            'Debug.WriteLine("new population created")
            If NormalizePopulation(newpopulation, reverse_flag) Then
                Return newpopulation
            End If
            Return Nothing
        End Function


        Public Sub Run(ByVal initial_population As IEnumerable(Of T), Optional ByVal reverse_flag As Boolean = False)
            If NormalizePopulation(initial_population, reverse_flag) Then
                _iteration = 0
                Dim population As List(Of T) = Nothing
                Do
                    population = NextPopulation(reverse_flag)
                    _iteration += 1
                Loop Until CanStop(_iteration, population)
            End If
        End Sub

#End Region

    End Class


End Namespace